/*
 * gen68k.h
 */


/*****************************************************************************
* Configuration                                                             */

/*
 * Address space, 24 = 24-bit, 32 = 32-bit
 * Default is 24
 */

#define GENITAL68K_ADDR 24


/*****************************************************************************
* Structure and data types                                                  */

typedef unsigned UINT32;        /* unsigned 32-bit */
typedef int INT32;              /* signed 32-bit */
typedef unsigned char BYTE;     /* unsigned 8-bit */

struct GENITAL68K_INSTRUCTIONREGION
{
        UINT32  base;
        UINT32  limit;
        BYTE    *offset;
};

struct GENITAL68K_DATAREGIONREAD
{
        UINT32  base;
        UINT32  limit;
        UINT32  (*handler)(UINT32);             /* address */
};

struct GENITAL68K_DATAREGIONWRITE
{
        UINT32  base;
        UINT32  limit;
        void    (*handler)(UINT32, UINT32);     /* address, data */
};
        
struct GENITAL68K_CPUCONTEXT
{
        struct GENITAL68K_INSTRUCTIONREGION     *fetch;
        struct GENITAL68K_DATAREGIONREAD        *read_byte, *read_word;
        struct GENITAL68K_DATAREGIONWRITE       *write_byte, *write_word;
        UINT32  a[8], d[8], sp, pc, sr;
        INT32   timer;
};


/*****************************************************************************
* Functions                                                                 */

int             Genital68KInit(struct GENITAL68K_CPUCONTEXT *);
void            Genital68KReset();
int             Genital68KRun();
struct GENITAL68K_CPUCONTEXT *Genital68KGetContext();
void            Genital68KSetContext(struct GENITAL68K_CPUCONTEXT *);
unsigned        Genital68KReadPC();
unsigned        Genital68KReadSP();
unsigned        Genital68KReadSR();
unsigned        Genital68KReadA(int);
unsigned        Genital68KReadD(int);


/*****************************************************************************
* Declarations                                                              */

extern struct GENITAL68K_CPUCONTEXT        genital68kcontext;
