/*                                                     
 * vdp.c: VDP emulation
 */

#include <stdio.h>
#include <stdlib.h>
#include "68k.h"
#include "genital.h"
#include "input.h"
#include "misc.h"
#include "vdp.h"
#include "vga.h"

struct vdp_struct       vdp;


/*
 * HV counter and Menacer management
 */

void VDPUpdateHVCounter(unsigned line)
{
        int     m;
//        if (!(vdp.reg[0] & 2))  /* if counter is stopped, don't update */
//                return;
//        if (line > 240)         /* maximum displayable line on PAL is 240 */
//                return;
        
        if (vdp.reg[12] & 1)    /* if h40 resolution... */
                vdp.hv_counter = (line << 8) | 0x9f;
        else                    /* h32 resolution... */
                vdp.hv_counter = (line << 8) | 0x7f;

        if (genesis.menacer)
        {
                if ((m = MouseClick()))
                {
                        if (m & 1)      /* left mouse button */
                                control.ctrl2 = 2;
                        else if (m & 2) /* right */
                                control.ctrl2 = genesis.mrbtn;
                }               
                else
                        control.ctrl2 = 0;
                vdp.hv_counter = MouseGetHV();
                CPUInterrupt(EXTINTERRUPT);
        }        
}

/*
 * Sprite order
 */

void VDPCalcSpriteOrder(int line)
{
        int             i = 0, j = 0, p = 0, pixlimit, x, y, w, h;
        unsigned        next = 0;

        if (genesis.spr_limit)  /* sprite limit and sprite masking */
        {
                pixlimit = (vdp.reg[12] & 1) ? 320 : 256;
                while (j < 256)
                {
                        x = GetWord(&vdp.sprite_addr[(next << 3) + 6]) & 0x1ff;
                        if (vdp.reg[12] & 6)    /* interlace 2 */
                                y = ((GetWord(&vdp.sprite_addr[next << 3]) & 0x3fe) >> 1) - 0x80;
                        else
                                y = (GetWord(&vdp.sprite_addr[next << 3]) & 0x1ff) - 0x80;
                        w = (((vdp.sprite_addr[(next << 3) + (2^1)] & 0xc) >> 2) + 1) * 8;
                        h = ((vdp.sprite_addr[(next << 3) + (2^1)] & 0x3) + 1) * 8;
       
                        if ((line >= y) && (line < (y + h)))
                        {
                                p += w;
                                if (p > pixlimit || !x)
                                {
                                        vdp.sprite_count = i + 1;
                                        return;
                                }
                                vdp.sprite_order[i] = next;
                                i++;                                
                        }
                        next = vdp.sprite_addr[((next << 3) + 3) ^ 1];
                        if (next == 0)
                        {
                                vdp.sprite_count = i;   /* # of sprites */
                                return;
                        }        
                        j++;
                }
                vdp.sprite_count = i + 1;
        }
        else
        {
                for (i = 0; i < 80; i++)
                {
                        vdp.sprite_order[i] = next;
                        /* ^1 for little endian */
                        next = vdp.sprite_addr[((next << 3) + 3) ^ 1];
                        if (next == 0)
                        {
                                vdp.sprite_count = i + 1;       /* sprite # */
                                return;
                        }
                }
                vdp.sprite_count = i + 1;
        }
}
                                              

/*
 * VDP RAM access
 */

inline unsigned ReadVDPRAMWord(unsigned addr)
{
        if (!vdp.access_mode)           /* access mode 0 = VRAM read */
                return GetWord(&vdp.vram[vdp.addr & 0xfffe]);
        else if (vdp.access_mode == 8)  /* mode 8 = CRAM read */
                return GetWord(&vdp.cram[vdp.addr]);
        else if (vdp.access_mode == 4)  /* mode 4 = VSRAM read */
                return GetWord(&vdp.vsram[vdp.addr & 0x7e]);
        return 0;
}

inline void WriteVDPRAMByte(unsigned addr, unsigned data)
{
        if (vdp.access_mode == 1)       /* access mode 1 = VRAM write */
                vdp.vram[addr] = data;
}
        
inline void WriteVDPRAMWord(unsigned addr, unsigned data)
{
        /* write word to VDP RAM (uses VDP context data for address and which
           RAM to write) */
        if (vdp.access_mode == 1)       /* access mode 1 = VRAM write */
        {
                if (addr & 1)
                        WORDSWAP(data);
                PutWord(&vdp.vram[addr & 0xfffe], data);
        }
        else if (vdp.access_mode == 3)  /* mode 3 = CRAM write */
        {
                PutWord(&vdp.cram[addr], data);
                SetPalette(vdp.addr, data);
        }
        else if (vdp.access_mode == 5)  /* mode 5 = VSRAM write */
                PutWord(&vdp.vsram[vdp.addr & 0x7e], data);
}

 
/*
 * DMA 
 */

void DoDMA()
{
        unsigned        i, len, addr;
        unsigned short  data;

        /* source address */
        addr = (vdp.reg[23] & 0x7f) << 17;
        addr += vdp.reg[22] << 9;
        addr += vdp.reg[21] << 1;

        len = (vdp.reg[20] << 8) + vdp.reg[19]; /* length */

        vdp.dma_mode = (vdp.reg[23] >> 6) & 3;

        switch (vdp.dma_mode)
        {                
                case 0: case 1:                 /* DMD1 = 0, DMD0 = 0 or 1,
                                                   memory-to-VRAM */                        
                        for (i = 0; i < len; i++)
                        {
                                data = MemReadWord(addr);       /* copy */
                                addr += 2;
                                WriteVDPRAMWord(vdp.addr, data);
                                vdp.addr += vdp.reg[15];
                        }
                        break;
                case 2:                         /* VRAM fill */
                        /* this is done on a write to the data port */
                        break;
                case 3:                         /* VRAM copy */
                        for (i = 0; i < len; i++)
                        {
                                data = ReadVDPRAMWord(addr);
                                addr += 2;
                                WriteVDPRAMWord(vdp.addr, data);
                                vdp.addr += vdp.reg[15];
                        }
                        break;
                default:
                        break;
        }
}

void DoDMAFill(unsigned data, int word) /* word = 1: use words, else bytes */
{
        int     i, len;

        len = (vdp.reg[20] << 8) + vdp.reg[19]; /* length */

        for (i = 0; i < len; i++)
                if (word)
                {
                        WriteVDPRAMWord(vdp.addr, data);
                        vdp.addr += vdp.reg[15];
                }
                else
                {
                        WriteVDPRAMByte(vdp.addr, data);
                        vdp.addr += vdp.reg[15];
                }

        vdp.dma_mode = 0;
}


/*
 * Registers
 */

void WindowH(int, unsigned);

void WriteReg(int i, unsigned data)
{
        vdp.reg[i] = data & 0xff;
        if (i == 12)    /* recalculate window width if screen changed */
                WindowH(17, vdp.reg[17]);                                       
}

void ScrollAAddr(int i, unsigned data)
{
        WriteReg(i, data);
        vdp.scroll_a_addr = &vdp.vram[(data & 0x38) << 10];
}

void WindowAddr(int i, unsigned data)
{
        WriteReg(i, data);
        if (vdp.reg[12] & 1)    /* bit 0 is 0 in H40 */
                vdp.window_addr = &vdp.vram[(data & 0x3e) << 10];
        else                    /* but not in H32 */
                vdp.window_addr = &vdp.vram[(data & 0x3e) << 10];
}

void ScrollBAddr(int i, unsigned data)
{
        WriteReg(i, data);
        vdp.scroll_b_addr = &vdp.vram[(data & 7) << 13];
}

void SpriteAddr(int i, unsigned data)
{
        WriteReg(i, data);
        vdp.sprite_addr = &vdp.vram[(data & 0x7f) << 9];
}

void HScrollAddr(int i, unsigned data)
{        
        WriteReg(i, data);
        vdp.hscroll_addr = &vdp.vram[(data & 0x3f) << 10];
}

/*
 * If H==0, it is shown for the full line _only_ when in V range. So if H==0
 * and V==0x9C, it will never be shown since it never gets in range. This
 * might have to be calculated in the DrawWindow() routine
 */

void WindowH(int i, unsigned data)
{
        int             scr_width = (vdp.reg[12] & 1) ? 40 : 32;
        unsigned        hd = (data & 0x80) >> 7, hs = (data & 0x1f) << 1;

        WriteReg(i, data);
        if (vdp.window_x >= scr_width)
                vdp.window_x = vdp.window_xlimit = 0;
            
        if (hd)                                 /* window on right */
        {
                vdp.window_x = hs;                  
                vdp.window_xlimit = scr_width - 1;
        }
        else                                    /* window on left */
        {
                vdp.window_x = 0;
                vdp.window_xlimit = hs;
                if (vdp.window_xlimit == 0)
                        vdp.window_xlimit = scr_width - 1;
        }
}

void WindowV(int i, unsigned data)
{
        unsigned        vd = (data & 0x80) >> 7, vs = data & 0x1f;

        WriteReg(i, data);
        if (vd)                                 /* window on lower side */
        {
                vdp.window_y = vs;
                if (vdp.window_y >= 28)
                        vdp.window_y = 0;
                vdp.window_ylimit = 28 - 1; 
        }
        else                                    /* window on upper side */
        {
                vdp.window_y = 0;
                if (!vs)
                        vdp.window_ylimit = 28 - 1;
                else
                        vdp.window_ylimit = vs - 1;
        }
}


/*
 * Ports
 */

unsigned ReadDataPortWord()
{
        unsigned        data;
        data = ReadVDPRAMWord(vdp.addr);
        vdp.addr += vdp.reg[15];
        return data;
}
        
void WriteDataPortByte(unsigned data)
{
        if (vdp.dma_mode == 2)  /* do VRAM fill */
                DoDMAFill(data, 0);
        else
        {
                WriteVDPRAMByte(vdp.addr, data);
                vdp.addr += vdp.reg[15];
        }
}

void WriteDataPortWord(unsigned data)
{
        if (vdp.dma_mode == 2)  /* do VRAM fill */
                DoDMAFill(data, 1);
        else 
        {
                WriteVDPRAMWord(vdp.addr, data);
                vdp.addr += vdp.reg[15];
        }
}
               
void DecodeCtrlPortCommand()
{
        /* access mode consists of cd0-cd3, cd4-cd5 are not needed */
        vdp.access_mode = (vdp.control_port >> 30) +
                          ((vdp.control_port >> 2) & 0xc);
        /* get address */
        switch (vdp.access_mode)
        {
                case 0:         /* VRAM read */
                        vdp.addr = (vdp.control_port >> 16) & 0x3fff;
                        vdp.addr += (vdp.control_port & 3) << 14;
                        break;
                case 8:         /* CRAM read */
                        vdp.addr = (vdp.control_port >> 16) & 0x7f;
                        break;
                case 4:         /* VSRAM read */
                        vdp.addr = (vdp.control_port >> 16) & 0x7f;
                        break;
                case 1:         /* VRAM write */
                        vdp.addr = (vdp.control_port >> 16) & 0x3fff;
                        vdp.addr += (vdp.control_port & 3) << 14;
                        break;
                case 3:         /* CRAM write */
                        vdp.addr = (vdp.control_port >> 16) & 0x7f;
                        break;
                case 5:         /* VSRAM write */
                        vdp.addr = (vdp.control_port >> 16) & 0x7f;
                        break;
                default:
                        break;
        }

        if (vdp.control_port & 0x80)    /* cd5 = 1 for DMA */
                DoDMA();
}

void WriteCtrlPortWord(unsigned data)
{
        typedef         void (*reg_call)(int, unsigned);
        reg_call        Reg[24] = { WriteReg, WriteReg, ScrollAAddr,
                                    WindowAddr, ScrollBAddr, SpriteAddr,
                                    WriteReg, WriteReg, WriteReg, WriteReg,
                                    WriteReg, WriteReg, WriteReg, HScrollAddr,
                                    WriteReg, WriteReg, WriteReg, WindowH,
                                    WindowV, WriteReg, WriteReg, WriteReg,
                                    WriteReg, WriteReg };

        if ((data >> 14) == 2)          /* write1 mode */
        {
                if (((data >> 8) & 0x1f) < 24)  /* make sure it is in range */
                        Reg[(data >> 8) & 0x1f]((data >> 8) & 0x1f, data);
        }
        else
        {
                if (vdp.ctrlport_mode == 0)     /* first word */
                {
                        vdp.control_port = data << 16;
                        vdp.ctrlport_mode++;    /* process word 2 next time */
                }
                else                            /* second word */
                {
                        vdp.control_port += data;
                        vdp.ctrlport_mode = 0;
                        DecodeCtrlPortCommand();        /* decode ctrl data */
                }
        }
}       


/*
 * Read/write handlers
 */

void VDPSetReg(int reg, unsigned data)
{
        typedef         void (*reg_call)(int, unsigned);
        reg_call        Reg[24] = { WriteReg, WriteReg, ScrollAAddr,
                                    WindowAddr, ScrollBAddr, SpriteAddr,
                                    WriteReg, WriteReg, WriteReg, WriteReg,
                                    WriteReg, WriteReg, WriteReg, HScrollAddr,
                                    WriteReg, WriteReg, WriteReg, WindowH,
                                    WindowV, WriteReg, WriteReg, WriteReg,
                                    WriteReg, WriteReg };

        if (reg < 24)   /* make sure it is in range */
                Reg[reg](reg, data);
}

void VDPWriteByte(unsigned addr, unsigned data)
{
        data &= 0xff;
       
        /* lower byte to data port */
        if (addr == 0xc00000 || addr == 0xc00001 || addr == 0xc00002 ||
            addr == 0xc00003)
                WriteDataPortByte(data);
}

void VDPWriteWord(unsigned addr, unsigned data)
{        
        data &= 0xffff;

        if (addr == 0xc00000 || addr == 0xc00002)       /* data port */
                WriteDataPortWord(data);
        else if (addr == 0xc00004 || addr == 0xc00006)  /* control port */
                WriteCtrlPortWord(data);
}

unsigned VDPReadByte(unsigned addr)
{
        if (addr == 0xc00004 || addr == 0xc00006)       /* status reg high */
        {
                vdp.status_register ^= 0x100;           /* FIFO kludge */
                vdp.status_register ^= 0x200;
                return (vdp.status_register >> 8);
        }
        if (addr == 0xc00005 || addr == 0xc00007)       /* status reg low */
        {
                vdp.status_register ^= 0x20;            /* collision kludge */
                vdp.status_register ^= 8;               /* DMA kludge */
                return (vdp.status_register & 0xff);
        }
        if (addr == 0xc00008)                           /* hv counter high */
                return (vdp.hv_counter >> 8);
        if (addr == 0xc00009)                           /* hv counter low */
                return (vdp.hv_counter & 0xff);
        return 0;
}

unsigned VDPReadWord(unsigned addr)
{
        if (addr == 0xc00000 || addr == 0xc00002)       /* data port */
                return ReadDataPortWord();
        if (addr == 0xc00004 || addr == 0xc00006)       /* status register */
        {
                vdp.status_register ^= 0x100;           /* FIFO kludge */
                vdp.status_register ^= 0x200;           
                vdp.status_register ^= 0x20;            /* collision kludge */
                vdp.status_register ^= 8;               /* DMA kludge */
                return vdp.status_register & 0xffff;
        }
        if (addr == 0xc00008)                           /* HV counter */
                return vdp.hv_counter & 0xffff;

        return 0;
}


/*
 * Interrupt management
 */

void VDPArmHIntTimer()
{
        vdp.hint_timer = vdp.reg[10];
}

void VDPHBlankStart()
{
        if (!genesis.hint)
                return;

        vdp.hint_timer--;
        if ((vdp.hint_timer < 0) && (vdp.reg[0] & 0x10))
        {
                vdp.status_register |= 2;       /* in hblank */
                CPUInterrupt(HINTERRUPT);
                VDPArmHIntTimer();
        }
}

void VDPHBlankEnd()
{
        vdp.status_register &= ~2;      /* no longer in hblank */
}

void VDPVBlankStart()
{                      
        if (vdp.reg[1] & 0x20)          /* if vertical interrupt allowed */
                CPUInterrupt(VINTERRUPT);
        vdp.status_register |= 0x80;    /* "V interrupt happened" */
        vdp.status_register |= 4;       /* in VBlank */
}

void VDPVBlankEnd()
{
        vdp.status_register &= ~0x80;   /* done with V interrupt */
        vdp.status_register &= ~4;      /* done w/ VBlank */
}


/*
 * Initialization
 */

void VDPInit()
{
        if ((vdp.vram = calloc(65536, sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 64KB RAM for " \
                        "VRAM\n");
                exit(1);
        }

        if ((vdp.cram = calloc(128, sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 128 bytes of " \
                        "RAM for CRAM\n");
                exit(1);
        }

        if ((vdp.vsram = calloc(80, sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 80 bytes of " \
                        "RAM for VSRAM\n");
                exit(1);
        }

        /* all register values are undefined at start up, pointers are
           initialized to prevent faults */

        vdp.scroll_a_addr = vdp.scroll_b_addr = vdp.hscroll_addr = vdp.vram;
        vdp.ctrlport_mode = vdp.dma_mode = vdp.hv_counter = 0;
        vdp.status_register = 0x200;    /* 0x200 for FIFO kludge to work */
}

void VDPDeinit()
{
        free(vdp.vram);
        free(vdp.cram);
}
